from numpy import deg2rad
import what3words
from typing import Tuple
from astropy.coordinates import SkyCoord
from astropy import units as u

# Connects to the what3words API using a public key (limited usage!!)
geocoder = what3words.Geocoder("6WEY7C3R")


def coords_to_words(*args, **kwargs) -> str:
    """
    Given RA, Dec of the object, return its "What 3 words" coordinate string.

    Args:
        ra (float or str): Right Ascension of the object [degrees or hms] AND
        dec (float or str): Declination of the object [degrees or dms], OR
        l (float): Galactic latitute [degrees] AND
        b (float): Galactic longitude [degrees], OR
        coord_string (str): 'XXhXXmXXs +XXdXXmXXs" string acceptable by AstroPy SkyCoord function. 
        format (str): coordinate format

    Returns:
        str: 3-word string delimited by periods corresponding to this coordinate.
    """
    # TODO
    # 1. Accept different types of coordinate inputs: hms:dms string, different coord systems,
    # or Astropy SkyCoord objects

    if ( (len(args) > 0 and isinstance(args[0], float)) or
       ("ra" in kwargs and isinstance(kwargs['ra'], float)) or
       ("l" in kwargs and isinstance(kwargs['l'], float))):
        c = SkyCoord(*args, **kwargs, unit=u.deg)
        print(c)
    else:
        c = SkyCoord(*args, **kwargs)
    c = c.transform_to('icrs')
    ra, dec = c.ra.deg, c.dec.deg

    # Check that the input is sensible
    if (ra < 0) or (ra > 360) or (dec < -90) or (dec > 90):
        raise ValueError("RA must be between 0 and 360, and Dec between -90 and 90!")
    
    # Convert ra to longitude-style (-180 < lon <180)
    if ra >= 180: ra -= 360

    # Search what3words API for the given latitude/longitude
    res = geocoder.convert_to_3wa(what3words.Coordinates(dec, ra))
    return res['words']


def words_to_coords(words:str) -> Tuple[float, float]:
    """
    Given three words, return the RA and dec coordinates of an object.

    Args:
        words (str): a string in the format "word1.word2.word3"

    Returns:
        tuple(float, float): Right ascension and declination of the object 
    """

    # Check that the input is sensible
    try:
        assert len(words.split('.')) == 3
    except:
        raise ValueError("Please pass the three word string in the format 'word1.word2.word3'.")

    # Search what3words API for the input coordinates
    res = geocoder.convert_to_coordinates(words)

    # If not found, raise an exception
    try:
        coords = res["coordinates"]
    except:
        raise ValueError("Error: these 3-word-coordinates are not found.")

    # Otherwise, return ra/dec
    ra, dec = coords["lng"], coords["lat"]
    if ra < 0: ra += 360
    return ra, dec



from astro3words import coords_to_words, words_to_coords
import what3words
import pytest
import numpy as np

def test_coords_to_words():
    """
    Tests conversion of (ra, dec) coordinates to 3 words 
    with some edge coordinate cases.
    """

    ra_list = [0, 90, 180, 360, 195.5]
    dec_list = [-90, 0, 89.9, 90]

    for ra in ra_list:
        for dec in dec_list:
            words = coords_to_words(ra, dec)
            assert isinstance(words, str)

def test_words_to_coords():
    """
    Tests backwards conversion: find 3 words for a list
    of coordinates, then make sure that these words
    map to the same coordinates within 1" precision.
    """

    ra_list = [0, 90, 180, 360, 195.5]
    dec_list = [-90, 0, 89.9, 90]
    precision = 1/3600

    for ra in ra_list:
        for dec in dec_list:
            words = coords_to_words(ra, dec)
            ra_out, dec_out = words_to_coords(words)

            # Calculate angular distance
            ra_diff = ra - ra_out 
            dec_diff = dec - dec_out
            if ra_diff > 180: ra_diff -= 360 # Wrap around between 0 and 360

            # Use small-angle approx of angular distance
            dist2 = ((ra_diff * np.pi / 180)**2 * 
                    np.cos(dec * np.pi/180)**2 +
                    (dec_diff*np.pi/180)**2)
            dist = np.sqrt(dist2)
            assert dist < precision

test_words_to_coords()



from setuptools import setup, find_packages
import numpy
import re

# borrowed from orbitize: https://github.com/sblunt/orbitize
def get_requires():
    reqs = []
    for line in open('requirements.txt', 'r').readlines():
        reqs.append(line)
    return reqs

# auto-updating version code stolen from RadVel
def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
                       open(project + '/__init__.py').read())
    return result.group(1)

setup(
    name='astro3words',
    version=get_property('__version__', 'astro3words'),
    description='astro3words - get a simple 3-word coordinate of your favorite object!',
    url='https://github.com/astro-nova/astro-3-words',
    author='',
    author_email='',
    license='MIT',
    packages=find_packages(),
    include_dirs=[numpy.get_include()],
    include_package_data = True,
    zip_safe=False,
    classifiers=[
        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',

        # Pick your license as you wish (should match "license" above)
        # 'License :: OSI Approved :: BSD License',

        'Programming Language :: Python :: 3',
        ],
    keywords='Astronomy Astrometry',
    install_requires=get_requires()
    )

